import torch
import torch.nn as nn

import pytorch_lightning as pl
import torchvision

import boris

from trainable import Trainable
from evaluatable import Evaluatable

from helpers import calculate_class_weights

import metrics


def configure_criterion(cfg, dataloader):

    if cfg['criterion']['name'] == 'cross-entropy':
        weight = calculate_class_weights(dataloader, cfg['num_classes'])
        return nn.CrossEntropyLoss(weight=weight)
    else:
        raise NotImplementedError()


def configure_dataloader(cfg, dataset, state=None, train=True):

    if train and state is not None:
        _, selected, _, _ = state
        sampler = torch.utils.data.SubsetRandomSampler(
            selected.nonzero()
        )
    else:
        sampler = None

    if train:
        dataloader = torch.utils.data.DataLoader(
            dataset,
            shuffle=True if sampler is None else False,
            batch_size=cfg['training']['batch_size'],
            sampler=sampler
        )
    else:
        dataloader = torch.utils.data.DataLoader(
            dataset,
            batch_size=cfg['evaluation']['batch_size'],
            num_workers=8
        )

    return dataloader


def configure_dataset(cfg, transform, train=True):

    if train:
        dataset_cfg = cfg['training']['data']
    else:
        dataset_cfg = cfg['evaluation']['data']

    dataset = boris.data.BorisDataset(**dataset_cfg, transform=transform)
    return dataset


def configure_embedding(cfg, data_cfg, transform_cfg):

    if 'resnet' in cfg['model']['name']:
        model = boris.models.ResNetSimCLR(**cfg['model'])
    else:
        raise NotImplementedError()

    criterion = boris.loss.NTXentLoss(**cfg['criterion'])
    optimizer = torch.optim.SGD(model.parameters(), **cfg['optimizer'])

    dataset = boris.data.BorisDataset(**data_cfg)

    collate_fn = boris.data.ImageCollateFunction(**cfg['collate'])
    dataloader = torch.utils.data.DataLoader(dataset,
                                             **cfg['loader'],
                                             collate_fn=collate_fn)

    embed_transform = torchvision.transforms.Compose([
        torchvision.transforms.Resize((transform_cfg['input_size'])),
        torchvision.transforms.ToTensor(),
        torchvision.transforms.Normalize(
            mean=[0.485, 0.456, 0.406],
            std=[0.229, 0.224, 0.225])
        ])
    embed_dataset = boris.data.BorisDataset(**data_cfg,
                                            transform=embed_transform)

    embedding_model = boris.embedding.SelfSupervisedEmbedding(
        model, criterion, optimizer, dataloader
    )
    return embedding_model, torch.utils.data.DataLoader(embed_dataset,
                                                        batch_size=100)


def configure_model(cfg):

    if 'resnet' in cfg['model']['name']:
        model = boris.models.ResNetGenerator(
            width=cfg['model']['width'],
            num_classes=cfg['num_classes']
        )
    else:
        raise NotImplementedError()

    return model


def configure_optimizer(cfg, model):

    if cfg['optimizer']['name'] == 'sgd':
        optimizer = torch.optim.SGD(
            model.parameters(),
            lr=cfg['optimizer']['lr'],
            weight_decay=cfg['optimizer']['weight_decay']
        )
    elif cfg['optimizer']['name'] == 'adam':
        optimizer = torch.optim.Adam(
            model.parameters(),
            lr=cfg['optimizer']['lr'],
            weight_decay=cfg['optimizer']['weight_decay']
        )
    else:
        raise NotImplementedError()

    return optimizer


def configure_scheduler(cfg, optimizer):
    if 'lr_decay' in cfg['optimizer']:
        lr_decay_step = cfg['optimizer']['lr_decay']
        scheduler = torch.optim.lr_scheduler.MultiStepLR(
            optimizer, milestones=lr_decay_step
        )
    else:
        scheduler = None
    return scheduler


def configure_transform(cfg, train=True):

    if train:
        transform_cfg = cfg['training']['transform']
    else:
        transform_cfg = cfg['evaluation']['transform']

    transforms = []

    if transform_cfg['input_size'] > 0:
        transforms.append(
            torchvision.transforms.Resize((transform_cfg['input_size']))
        )

    if 'flip' in transform_cfg.keys() and transform_cfg['flip']:
        transforms.append(
            torchvision.transforms.RandomHorizontalFlip()
        )

    transforms.append(torchvision.transforms.ToTensor())

    if 'normalize' in transform_cfg.keys() and transform_cfg['normalize']:
        transforms.append(
            torchvision.transforms.Normalize(
                mean=[0.485, 0.456, 0.406],
                std=[0.229, 0.224, 0.225]
            )
        )

    return torchvision.transforms.Compose(transforms)


def configure_trainable(cfg, state, device=None):

    model = configure_model(cfg)
    optimizer = configure_optimizer(cfg, model)
    scheduler = configure_scheduler(cfg, optimizer)

    transform = configure_transform(cfg, train=True)
    dataset = configure_dataset(cfg, transform, train=True)
    dataloader = configure_dataloader(cfg, dataset, state=state, train=True)
    criterion = configure_criterion(cfg, dataloader)

    trainable = Trainable(model, criterion, optimizer, dataloader, scheduler)
    trainer = pl.Trainer(**cfg['training']['trainer'])

    return trainable, trainer


def configure_evaluatable(cfg):

    num_classes = cfg['num_classes']

    transform = configure_transform(cfg, train=False)
    dataset = configure_dataset(cfg, transform, train=False)
    dataloader = configure_dataloader(cfg, dataset, train=False)
    eval_metrics = []
    for metric in cfg['evaluation']['metrics']:

        # classification
        if metric == 'classification-per-class-accuracy':
            eval_metric = [
                metrics.ClassificationPerClassAccuracy(i, num_classes)
                for i in range(num_classes)
            ]
        elif metric == 'classification-top-1-accuracy':
            eval_metric = [metrics.ClassificationTop1Accuracy()]
        elif metric == 'classification-top-5-accuracy':
            eval_metric = [metrics.ClassificationTop5Accuracy()]
        elif metric == 'classification-mean-precision':
            eval_metric = [metrics.ClassificationMeanPrecision(num_classes)]
        elif metric == 'classification-mean-recall':
            eval_metric = [metrics.ClassificationMeanRecall(num_classes)]
        # other
        elif metric == 'segmentation-TODO':
            raise NotImplementedError()
        # none of the above
        else:
            raise ValueError('Illegal metric: {}'.format(eval_metric))

        eval_metrics = eval_metrics + eval_metric

    evaluatable = Evaluatable(dataloader, eval_metrics)
    return evaluatable



import pytorch_lightning.core.lightning as lightning


class Trainable(lightning.LightningModule):

    def __init__(self,
                 model,
                 criterion,
                 optimizer,
                 dataloader,
                 scheduler=None):

        super(Trainable, self).__init__()
        self.model = model
        self.criterion = criterion
        self.optimizer = optimizer
        self.dataloader = dataloader
        self.scheduler = scheduler

    def forward(self, x):
        return self.model(x)

    def training_step(self, batch, batch_idx):
        x, y, _ = batch
        y_hat = self(x)
        loss_ = self.criterion(y_hat, y)
        tensorboard_logs = {'train_loss': loss_}
        return {'loss': loss_, 'log': tensorboard_logs}

    def configure_optimizers(self):
        if self.scheduler is None:
            return self.optimizer
        else:
            return [self.optimizer], [self.scheduler]

    def train_dataloader(self):
        return self.dataloader



import pandas as pd
import numpy as np


class Logger(object):

    def __init__(self, name):
        self.name = 'benchmark.' + name
        self.rows = []

    def add_row(self, row):
        self.rows.append(row)

    def to_pandas(self, header):
        df = pd.DataFrame(columns=header)
        for row in self.rows:
            df = df.append(
                dict(zip(header, row)), ignore_index=True
            )
        return df

    def backup(self, header):
        df = self.to_pandas(header)
        df.to_csv(self.name + '.csv')


def dump_embeddings(embeddings, labels, id_):

    embeddings = np.array(embeddings).astype(float)
    labels = np.array(labels).astype(float).reshape(-1, 1)

    data = np.concatenate((embeddings, labels), 1)
    np.savetxt('embeddings-' + str(id_) + '.csv', data)



import seaborn
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker

import pandas as pd
import numpy as np
import umap

import shutil
import glob
import os


REPORT_DIR = './reports/'
OUTPUT_DIR = './outputs/'


def _find_lower_lim(x):
    """Find the next multiple of 10 which is lower than x

    """
    x = int(np.floor((x * 100)))
    while x % 10 > 0:
        x -= 1
    return x / 100.


def _find_upper_lim(x):
    """Find the next multiple of 10 which is larger than x

    """
    x = int(np.ceil(x * 100))
    while x % 10 > 0:
        x += 1
    return x / 100.


def make_box_plot(ax, df, metric, sample_percentage, m, M):
    """Make a box plot which shows the metric for all strategies
       at the given percentage of the data.

    Args:
        ax: matplotlib axes
        df: Pandas dataframe containing the results (see log.py -> Logger)
        metric: The metric to show (e.g. classification-top-1-accuracy)
        sample_percentage: At which percentage of the data to plot
            (e.g. 0.1 for 10%)
        m: lower bound for the x-axis
        M: upper bound for the x-axis

    """

    df_copy = df.loc[abs(df['sample_percentage'] - sample_percentage) < 1e-2]

    seaborn.boxplot(x=metric, y='embedding_model', hue='sampling_strategy',
                    data=df_copy, palette='vlag', ax=ax, width=0.25)

    ax.xaxis.grid(True)
    ax.yaxis.grid(False)
    ax.set(ylabel="")
    ax.set(title=str(int(100 * sample_percentage)) + '% of the Data')

    ax.set(xlim=[m, M])
    ax.xaxis.set_major_locator(ticker.MultipleLocator(0.05))
    ax.legend(loc='lower right')


def make_line_plot(ax, df, metric, embedding_model):
    """Make a line plot which shows the metric for all strategies
       at all percentages of the data.

    Args:
        ax: matplotlib axes
        df: Pandas dataframe containing the results (see log.py -> Logger)
        metric: The metric to show (e.g. classification-top-1-accuracy)
        embedding_model: The model for which the plots are made

    """

    df_copy = df.loc[df['embedding_model'] == embedding_model]

    palette = 'colorblind'
    seaborn.lineplot(x='sample_percentage', y=metric, hue='sampling_strategy',
                     data=df_copy, dashes=True, LineWidth=2.,
                     palette=palette, err_style='bars')

    ax.yaxis.grid(True)
    ax.xaxis.grid(True)
    ax.set(xlim=[0.1, 1.1])
    ax.set(ylim=[0, 1])
    ax.set(aspect='equal')
    ax.yaxis.set_major_locator(ticker.MultipleLocator(0.1))
    ax.xaxis.set_major_locator(ticker.MultipleLocator(0.1))

    ax.legend(loc='lower right')


def make_scatter_plot(X, y, title):
    """Make a scatter plot of the embeddings in X

    Args:
        X: Embeddings (N x d) numpy array
        y: Labels for coloring
        title: Title for the plot

    """

    trans = umap.UMAP().fit(X)

    seaborn.set_style("darkgrid")

    f, ax = plt.subplots(figsize=(10, 10))
    f.suptitle(title.upper(), x=0., y=.93, horizontalalignment='left')

    seaborn.scatterplot(x=trans.embedding_[:, 0], y=trans.embedding_[:, 1],
                        hue=y.astype(int), legend='full',
                        palette='tab20b')

    ax.set(aspect='equal')

    m = trans.embedding_[:, :2].min()
    M = trans.embedding_[:, :2].max()

    ax.set(xlim=(m-1, M+1))
    ax.set(ylim=(m-1, M+1))

    return f


def make_box_plots(df, metric):
    """Make all box plots for the specified metric

    """

    percentages = df['sample_percentage'].unique()
    n_percentages = len(percentages)
    n_embeddings = len(df['embedding_model'].unique())
    n_samplers = len(df['sampling_strategy'].unique())
    f, axs = plt.subplots(n_percentages, figsize=(10, n_percentages *
                                                  (n_embeddings *
                                                   n_samplers)))

    m = _find_lower_lim(df[metric].min())
    M = _find_upper_lim(df[metric].max())

    if n_percentages < 2:
        axs = [axs]

    for i, percentage in enumerate(percentages):
        make_box_plot(axs[i], df, metric, percentage, m, M)

    f.tight_layout(rect=[0, 0, 1, 0.95])
    f.suptitle(metric.upper(), x=0., y=.98, horizontalalignment='left')

    return f


def make_line_plots(df, metric):
    """Make all line plots for the specified metric

    """

    n_embeddings = len(df['embedding_model'].unique())
    f, axs = plt.subplots(n_embeddings, figsize=(10, 10
                                                 * n_embeddings))

    if n_embeddings < 2:
        axs = [axs]

    for i, model in enumerate(df['embedding_model'].unique()):
        make_line_plot(axs[i], df, metric, model)

    f.tight_layout(rect=[0, 0, 1, 0.9])
    f.suptitle(metric.upper(), x=0., y=.93, horizontalalignment='left')

    return f


def make_plots(df, Xs, path=None):
    """Make all plots and save them if path is specified

    Args:
        df: pandas dataframe which stores the results from a benchmark run
        Xs: list of tuples (X, filename)
            where X is the embedding stored in filename
        path: path to the directory where the plots are to be stored

    """

    seaborn.set_style('darkgrid')

    # box plots
    box_plots = []
    for column in df.columns[5:]:
        f = make_box_plots(df, column)
        box_plots.append(f)

    # save box plots
    if path is not None:
        for f in box_plots:
            name = f._suptitle.get_text()
            name = 'BOX-' + name
            tmp_path = os.path.join(path, name)
            tmp_path = tmp_path + '.pdf'
            f.savefig(tmp_path, format='pdf')
            plt.close(f)

    # line plots
    line_plots = []
    for column in df.columns[5:]:
        f = make_line_plots(df, column)
        line_plots.append(f)

    # save line plots
    if path is not None:
        for f in line_plots:
            name = f._suptitle.get_text()
            name = 'LINE-' + name
            tmp_path = os.path.join(path, name)
            tmp_path = tmp_path + '.pdf'
            f.savefig(tmp_path, format='pdf')
            plt.close(f)

    # make scatter plots
    scatter_plots = []
    for X, filename in Xs:
        f = make_scatter_plot(X[:, :-1], X[:, -1], filename)
        scatter_plots.append(f)

    # save scatter plots
    if path is not None:
        for f in scatter_plots:
            name = f._suptitle.get_text()
            name = 'SCATTER-' + name
            tmp_path = os.path.join(path, name)
            tmp_path = tmp_path + '.pdf'
            f.savefig(tmp_path, format='pdf')
            plt.close(f)


if __name__ == '__main__':

    benchmark_dirs = glob.glob(
        OUTPUT_DIR + '*-*-*/*-*-*/'
    )

    # create report directory
    if not os.path.exists(REPORT_DIR):
        os.mkdir(REPORT_DIR)

    for benchmark_dir in benchmark_dirs:

        report_dir = os.path.join(
            REPORT_DIR,
            '/'.join(benchmark_dir.split('/')[2:])
        )

        if not os.path.exists(report_dir):
            os.makedirs(report_dir)

        print('Processing: ' + benchmark_dir)
        print('Report to : ' + report_dir)

        # read log file
        log_file = glob.glob(benchmark_dir + 'benchmark.*.csv')[-1]
        shutil.copy2(log_file, report_dir)
        df = pd.read_csv(log_file)

        # read embeddings
        Xs = []
        embedding_files = glob.glob(benchmark_dir + 'embeddings-*.csv')
        for embedding_file in embedding_files:
            shutil.copy2(embedding_file, report_dir)
            X = np.loadtxt(embedding_file)
            name = os.path.basename(embedding_file)
            Xs.append((X, name))

        # make plots
        plot_dir = os.path.join(report_dir + 'plots')
        if not os.path.exists(plot_dir):
            os.mkdir(plot_dir)

        make_plots(df, Xs, path=plot_dir)



import torch

from helpers import confusion_matrix


class BaseMetric(object):

    def __init__(self, identifier):
        self._identifier = identifier
        self._result = None

    @property
    def identifier(self):
        return self._identifier

    @property
    def result(self):
        return self._result.detach().cpu().item()

    @result.setter
    def result(self, value):
        if self._result is None:
            self._result = value
        else:
            self._result += value


class ClassificationTop1Accuracy(BaseMetric):

    def __init__(self):
        super(ClassificationTop1Accuracy, self).__init__(
            'classification-top-1-accuracy'
        )

    def __call__(self, output, target, N):
        _, argmax = output.max(-1)
        self.result = (argmax == target).float().mean() / N


class ClassificationTop5Accuracy(BaseMetric):

    def __init__(self):
        super(ClassificationTop5Accuracy, self).__init__(
             'classification-top-5-accuracy'
        )

    def __call__(self, output, target, N):
        _, argsort = torch.sort(output, -1, descending=True)
        accuracy = torch.zeros(len(output)).to(output.device)
        for i in range(5):
            accuracy += (argsort[:, i] == target)
        self.result = accuracy.mean() / N


class ClassificationMeanPrecision(BaseMetric):

    def __init__(self, n_classes):
        super(ClassificationMeanPrecision, self).__init__(
            'classification-mean-precision'
        )
        self.n_classes = n_classes

    def __call__(self, output, target, N):
        C = confusion_matrix(output, target, self.n_classes)
        precision = torch.diag(C) / torch.sum(C, 0)
        precision[precision != precision] = 0.
        self.result = precision.mean() / N


class ClassificationMeanRecall(BaseMetric):

    def __init__(self, n_classes):
        super(ClassificationMeanRecall, self).__init__(
            'classification-mean-recall'
        )
        self.n_classes = n_classes

    def __call__(self, output, target, N):
        C = confusion_matrix(output, target, self.n_classes)
        recall = torch.diag(C) / torch.sum(C, 1)
        recall[recall != recall] = 0.
        self.result = recall.mean() / N


class ClassificationPerClassAccuracy(BaseMetric):

    def __init__(self, class_id, n_classes):
        super(ClassificationPerClassAccuracy, self).__init__(
            'classification-per-class-accuracy-' + str(class_id)
        )
        self.class_id = class_id
        self.n_classes = n_classes

    def __call__(self, output, target, N):
        C = confusion_matrix(output, target, self.n_classes)
        C = C / C.sum(axis=1)[:, None]
        C[C != C] = 0.
        self.result = torch.diag(C)[self.class_id] / N



import setuptools
import sys


if __name__ == '__main__':

    with open('README.md') as f:
        long_description = f.read()

    name = 'borisml'
    version = '1.0.0'

    author = 'Philipp Wirth & Igor Susmelj'
    author_email = 'philipp@lightly.ai'
    description = "This package has been deprecated, please install lightly."

    # keep entry points but display a message indicating that borisml has
    # migrated to lightly
    entry_points = {
        "console_scripts": [
            "boris-train = boris.cli.train_cli:entry",
            "boris-embed = boris.cli.embed_cli:entry",
            "boris-magic = boris.cli.boris_cli:entry",
            "boris-upload = boris.cli.upload_cli:entry",
            "boris-download = boris.cli.download_cli:entry",
        ]
    }

    # require lightly package which replaces borisml
    install_requires = ['lightly']

    project_urls = {
        'Python Package': 'https://pypi.org/project/lightly/',
        'Documentation': 'https://lightly.readthedocs.io',
        'Lightly': 'https://www.lightly.ai',
    }

    classifiers = [
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License"
    ]

    setuptools.setup(
        name=name,
        version=version,
        author=author,
        author_email=author_email,
        description=description,
        entry_points=entry_points,
        license='MIT',
        long_description=long_description,
        long_description_content_type='text/markdown',
        install_requires=install_requires,
        packages=setuptools.find_packages(),
        classifiers=classifiers,
        include_package_data=True,
        project_urls=project_urls,
    )




DOCS = 'https://lightly.readthedocs.io'


def notify(task='train'):
    msg = ''
    msg += 'The borisml Python package has been renamed to lightly.\n'
    msg += f'Instead of boris-{task}, please try out lightly-{task}.\n'
    msg += f'Check out the documentation: {DOCS}'
    print(msg)
